import pytest

from pychores.configmodule import Config, Production


class TestConfigModule:
    def test_get_dev_config(self):
        config = Config.get_config("development")
        assert config.DEBUG

    def test_get_prod_config_raise_if_no_db_uri_is_set(self):
        with pytest.raises(ValueError):
            Config.get_config("production")

    def test_get_prod_config_raise_if_secret_is_unchanged(self, monkeypatch):
        old_db_uri = Config.SQLALCHEMY_DATABASE_URI
        Config.SQLALCHEMY_DATABASE_URI = "fake_uri"
        with pytest.raises(ValueError):
            Config.get_config("production")
        Config.SQLALCHEMY_DATABASE_URI = old_db_uri

    def test_default_config_is_prod_config(self):
        with pytest.raises(ValueError):
            config = Config.get_config("anything")
            assert isinstance(config, Production)



"""
A flask app to help you manage your recurent tasks

I tried to use flak, sqlalchemy and pytest as cleanly as possible
This app is also a test about publishing a package using flit
"""

from flask import Flask, jsonify
from flask_cors import CORS
from flask_jwt_extended import JWTManager

from ._version import version
from .configmodule import Config
from .model import DeferedSession

jwt = JWTManager()

__version__ = version


def respond_success(title, content):
    response_object = {
        "status": "success",
        title: content,
    }
    return jsonify(response_object)


def create_app(environment):
    app = Flask(__name__, static_folder="app", static_url_path="/")
    app.config.from_object(Config.get_config(environment))
    jwt.init_app(app)
    app.db_session = DeferedSession.get_session_local(
        app.config["SQLALCHEMY_DATABASE_URI"]
    )

    CORS(app, resources={r"/*": {"origins": "*"}})

    # sanity check route
    @app.route("/infos", methods=["GET"])
    def api_version():
        return respond_success("version", __version__)

    from .routes import auth, chores, tasks

    app.register_blueprint(auth.bp)
    app.register_blueprint(tasks.bp)
    app.register_blueprint(chores.bp)

    @app.route("/", defaults={"path": ""})
    @app.route("/<path:path>")
    def catch_all(path):
        return app.send_static_file("index.html")

    return app



import pytest

from pychores.domain.entity.user import User
from pychores.domain.use_cases.sign_user_in import IGetUserRepo, SignUserIn


class DumyGetUserRepo(IGetUserRepo):
    def get_user(self, username: str) -> User:
        return User(email="fake@example.com", username="fake", password="hunter2")


def dumy_password_checker(crypted_pass: str, raw_pass: str) -> bool:
    return crypted_pass == raw_pass


class TestSignUserIn:
    def test_wrong_password_should_raise_value_error(self):
        uc = SignUserIn(DumyGetUserRepo(), dumy_password_checker)
        with pytest.raises(ValueError):
            uc.execute({"username": "fake", "password": "bad pass"})



from abc import ABC, abstractmethod

from pychores.domain.entity.chore import Chore


class IGetChores(ABC):
    @abstractmethod
    def get_chores(self, username: str) -> list[Chore]:
        """return a list of Chore for username"""


class GetUserChores:
    def __init__(self, repo: IGetChores):
        self.repo = repo

    def execute(self, username: str) -> list[Chore]:
        chores = self.repo.get_chores(username)
        return sorted(chores, key=lambda x: x.next_execution)



from datetime import date

from flask import Blueprint, abort, current_app, jsonify, request, url_for
from flask_cors import CORS
from flask_jwt_extended import get_jwt_identity, jwt_required

from pychores.adapter.repository.sqla.chore import ChoreRepo
from pychores.adapter.repository.sqla.task import TaskRepo
from pychores.domain.use_cases.add_task import AddTask
from pychores.domain.use_cases.create_chore import CreateChore, InvalidChorePayload
from pychores.domain.use_cases.delete_chore import DeleteChore
from pychores.domain.use_cases.edit_chore import EditChore
from pychores.domain.use_cases.get_chore import GetChore
from pychores.domain.use_cases.get_user_chores import GetUserChores

from .. import respond_success
from ..model import Chore, User
from .tasks import respond_task

bp = Blueprint("chores", __name__, url_prefix="/api")
CORS(bp, resources={r"/*": {"origins": "*"}})


def serialize_chore(chore):
    return {
        "id": chore.id,
        "name": chore.name,
        "description": chore.description,
        "period": chore.period,
        "last_execution_date": chore.last_execution,
        "next_execution_date": chore.next_execution,
        "is_overdue": chore.is_overdue,
    }


def respond_chore(chore):
    response = jsonify(serialize_chore(chore))
    response.headers["Location"] = url_for(".chore", chore_id=chore.id)
    return response


@bp.route("/chore", methods=["GET", "POST"])
@jwt_required()
def chores():
    current_username = get_jwt_identity()

    if request.method == "POST":
        post_data = request.get_json()
        uc = CreateChore(ChoreRepo(current_app.db_session), date.today)
        try:
            chore = uc.execute(username=current_username, payload=post_data)
        except InvalidChorePayload:
            return "Invalid chore payload", 400
        return respond_chore(chore)

    # get

    uc = GetUserChores(ChoreRepo(current_app.db_session))
    chores = uc.execute(current_username)
    serialized_chores = [serialize_chore(chore) for chore in chores]
    return respond_success("chores", serialized_chores)


@bp.route("/chore/<int:chore_id>", methods=["GET", "PUT", "DELETE", "POST"])
@jwt_required()
def chore(chore_id):
    username = get_jwt_identity()
    current_user = (
        current_app.db_session.query(User)
        .filter_by(username=get_jwt_identity())
        .first()
    )
    chore = current_app.db_session.get(Chore, chore_id)
    if chore.user_id != current_user.id:
        abort(401)
    if request.method == "PUT":
        uc = EditChore(ChoreRepo(current_app.db_session))
        chore = uc.execute(
            username=username, chore_id=chore_id, payload=request.get_json()
        )
        return respond_chore(chore)

    if request.method == "DELETE":
        uc = DeleteChore(ChoreRepo(current_app.db_session))
        uc.execute(username, chore_id=chore_id)
        return "", 204
    if request.method == "POST":
        uc = AddTask(
            task_repo=TaskRepo(current_app.db_session),
            chore_repo=ChoreRepo(current_app.db_session),
        )
        task = uc.execute(username, chore_id, payload=request.get_json() or dict())
        return respond_task(task)
    #  get
    uc = GetChore(ChoreRepo(current_app.db_session))
    chore = uc.execute(username=username, chore_id=chore_id)
    return respond_chore(chore)



from flask import Blueprint, current_app, jsonify, request
from flask_cors import CORS
from flask_jwt_extended import create_access_token

from pychores.adapter.provider.hash import check_provider, hash_provider
from pychores.adapter.repository.sqla.user import UserRepo
from pychores.domain.use_cases.create_user import CreateUser
from pychores.domain.use_cases.sign_user_in import SignUserIn

bp = Blueprint("auth", __name__, url_prefix="/api")
CORS(bp, resources={r"/*": {"origins": "*"}})


def serialize_user(user, access_token=None):
    return {
        "username": user.username,
        "email": user.email,
        "id": user.id,
        "access_token": access_token,
    }


@bp.route("/auth/signup", methods=["POST"])
def create_user():
    post_data = request.get_json()
    uc = CreateUser(UserRepo(current_app.db_session), hash_provider)
    user = uc.execute(post_data)
    access_token = create_access_token(identity=user.username)
    return jsonify(serialize_user(user, access_token))


@bp.route("/auth/signin", methods=["POST"])
def signin():
    post_data = request.get_json()
    uc = SignUserIn(UserRepo(current_app.db_session), check_provider)
    user = uc.execute(post_data)
    access_token = create_access_token(identity=user.username)
    return jsonify(serialize_user(user, access_token))



from abc import ABC, abstractmethod

from pychores.domain.entity.chore import Chore
from pychores.domain.entity.task import Task


class IGetChore(ABC):
    @abstractmethod
    def get_chore(self, username: str, chore_id: int) -> Chore:
        """Return a chore owned by username"""


class IAddTask(ABC):
    @abstractmethod
    def save_task(self, task: Task) -> Task:
        """Add a new task for a chore"""


class AddTask:
    def __init__(self, chore_repo: IGetChore, task_repo: IAddTask):
        self.task_repo = task_repo
        self.chore_repo = chore_repo

    def execute(self, username: str, chore_id: int, payload: dict) -> Task:
        chore = self.chore_repo.get_chore(username=username, chore_id=chore_id)
        task = Task(
            chore=chore,
            execution_date=payload.get("execution_date", chore.current_date),
        )
        task = self.task_repo.save_task(task)
        return task



from sqlalchemy import func

from pychores.domain.entity.chore import Chore
from pychores.domain.use_cases.create_chore import ICreateChore
from pychores.domain.use_cases.get_chore import IGetChore
from pychores.domain.use_cases.get_user_chores import IGetChores
from pychores.model import Chore as DbChore
from pychores.model import Task as DbTask
from pychores.model import User as DbUser

from .builder import build_chore


class ChoreRepo(IGetChores, ICreateChore, IGetChore):
    def __init__(self, session):
        self.session = session

    def get_chores(self, username: str) -> list[Chore]:
        last_tasks = (
            self.session.query(
                DbTask.chore_id,
                func.max(DbTask.execution_date).label("last_execution"),
            )
            .group_by(DbTask.chore_id)
            .subquery()
        )
        db_chores = (
            self.session.query(DbChore)
            .join(DbUser)
            .outerjoin(last_tasks, last_tasks.c.chore_id == Chore.id)
            .filter(DbUser.username == username)
        )

        return [build_chore(c) for c in db_chores]

    def get_chore(self, username: str, chore_id: int) -> Chore:
        db_chore = self.session.get(DbChore, chore_id)
        assert db_chore.user.username == username
        return build_chore(db_chore)

    def save_new_chore(self, username: str, chore: Chore) -> Chore:
        db_user = self.session.query(DbUser).filter_by(username=username).one()
        db_chore = DbChore(
            name=chore.name, description=chore.description, period=chore.period
        )
        db_user.chores.append(db_chore)
        self.session.commit()
        return build_chore(db_chore)

    def save_chore(self, chore: Chore) -> Chore:
        db_chore = self.session.get(DbChore, chore.id)
        db_chore.name = chore.name
        db_chore.description = chore.description
        db_chore.period = chore.period
        self.session.commit()
        return chore

    def delete_chore(self, chore: Chore):
        db_chore = self.session.get(DbChore, chore.id)
        self.session.delete(db_chore)
        self.session.commit()



from dataclasses import field
from datetime import date, timedelta

from pydantic.dataclasses import dataclass


@dataclass
class Chore:
    name: str
    description: str
    period: int
    current_date: date
    last_execution: date | None
    id: int | None = None
    next_execution: date = field(init=False)
    is_overdue: bool = field(init=False)

    def __post_init__(self):
        self.next_execution = (
            self.last_execution + timedelta(days=self.period)
            if self.last_execution
            else date.today()
        )
        self.is_overdue = self.next_execution < self.current_date



from abc import ABC, abstractmethod

from pychores.domain.entity.chore import Chore


class IDeleteChore(ABC):
    @abstractmethod
    def get_chore(self, username: str, chore_id: int) -> Chore:
        """Return a Chore owned by username"""

    @abstractmethod
    def delete_chore(self, chore: Chore):
        """Delete `chore`"""


class DeleteChore:
    def __init__(self, repo: IDeleteChore):
        self.repo = repo

    def execute(self, username: str, chore_id: int):
        Chore = self.repo.get_chore(username=username, chore_id=chore_id)
        self.repo.delete_chore(Chore)



from pychores.domain.entity.user import User
from pychores.domain.use_cases.create_user import ISaveUserRepo
from pychores.domain.use_cases.sign_user_in import IGetUserRepo
from pychores.model import User as DbUser


class UserRepo(ISaveUserRepo, IGetUserRepo):
    def __init__(self, session):
        self.session = session

    def save_user(self, user: User):
        db_user = DbUser(
            username=user.username, email=user.email, password=user.password
        )

        self.session.add(db_user)
        self.session.commit()

    def get_user(self, username: str) -> User:
        db_user = self.session.query(DbUser).filter_by(username=username).first()
        return User(
            username=db_user.username,
            email=db_user.email,
            password=db_user.password,
        )



from datetime import date

from sqlalchemy import ForeignKey, String, create_engine
from sqlalchemy.orm import (
    DeclarativeBase,
    Mapped,
    mapped_column,
    relationship,
    scoped_session,
    sessionmaker,
)


class DeferedSession:
    db_uri = None
    session_local = None

    @classmethod
    def get_session_local(cls, db_uri=None):
        if db_uri is None and cls.db_uri is None:
            raise ValueError("No engine !!")
        elif db_uri is not None and cls.db_uri is not None and cls.db_uri != db_uri:
            raise ValueError(f"different engine !!! {cls.db_uri=}, {db_uri=}")
        if cls.db_uri is None:
            cls.db_uri = db_uri
            engine = create_engine(db_uri)
            cls.session_local = scoped_session(sessionmaker(engine))

        return cls.session_local


class Base(DeclarativeBase):
    pass


class Chore(Base):
    __tablename__ = "chore"
    id: Mapped[int] = mapped_column(primary_key=True)
    user_id: Mapped[int] = mapped_column(ForeignKey("user.id"))
    user: Mapped["User"] = relationship("User", back_populates="chores")
    name: Mapped[str] = mapped_column(unique=True)
    description: Mapped[str]
    period: Mapped[int]
    tasks: Mapped[list["Task"]] = relationship(
        "Task", back_populates="chore", cascade="delete, all"
    )


class Task(Base):
    __tablename__ = "task"
    id: Mapped[int] = mapped_column(primary_key=True)
    chore_id: Mapped[int] = mapped_column(ForeignKey("chore.id"))
    chore: Mapped[Chore] = relationship("Chore", back_populates="tasks")
    execution_date: Mapped[date]


class User(Base):
    __tablename__ = "user"
    id: Mapped[int] = mapped_column(primary_key=True)
    username: Mapped[str] = mapped_column(String(50), unique=True)
    password: Mapped[str] = mapped_column(String(60))
    email: Mapped[str] = mapped_column(String(50))
    chores: Mapped[list[Chore]] = relationship(Chore, back_populates="user")



"""Add user table

Revision ID: 4f016505c7c3
Revises: 615f2c3df9bd
Create Date: 2020-07-02 14:03:35.515584

"""
import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision = "4f016505c7c3"
down_revision = "615f2c3df9bd"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "user",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("username", sa.String(length=50), nullable=False),
        sa.Column("password", sa.String(length=60), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("username"),
    )
    with op.batch_alter_table("chore", schema=None) as batch_op:
        batch_op.alter_column("description", existing_type=sa.VARCHAR(), nullable=True)
        batch_op.alter_column("period", existing_type=sa.INTEGER(), nullable=True)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column("chore", "period", existing_type=sa.INTEGER(), nullable=False)
    op.alter_column("chore", "description", existing_type=sa.VARCHAR(), nullable=False)
    op.drop_table("user")
    # ### end Alembic commands ###



import pytest

from pychores.model import DeferedSession


class TestDeferdSession:
    def test_should_raise_if_no_engine_has_ever_been_set(self):
        session_value = DeferedSession.session_local
        uri_value = DeferedSession.db_uri
        DeferedSession.session_local = None
        DeferedSession.db_uri = None
        with pytest.raises(ValueError):
            DeferedSession.get_session_local()
        DeferedSession.session_local = session_value
        DeferedSession.db_uri = uri_value

    def test_should_raise_if_different_engine_is_requested(self):
        with pytest.raises(ValueError):
            DeferedSession.get_session_local(db_uri="dumy uri")
